import unittest
from typing import NamedTuple

import diffrax
import matplotlib.pyplot as plt

from jax_chmc.rattle import Rattle
import jax.numpy as jnp
import jax
from diffrax import ODETerm, SaveAt
import numpy as np
from diffrax import NewtonNonlinearSolver
from diffrax.custom_types import Bool, DenseInfo, PyTree, Scalar,Array
import jax.tree_util as jtu
from equinox.internal import ω

class RattleTestCase(unittest.TestCase):


    def test_circle(self):
        def constrain(q):
            return jnp.sqrt(jnp.sum(q**2, keepdims=True))-1.

        rat = Rattle(nonlinear_solver=NewtonNonlinearSolver(rtol=1e-4, atol=1e-6),constrain=constrain)


        def H(p, q):
            del q
            return p @ p.T / 2.

        # V = p^2/2m m=1, v=1

        # (p,q)->(0,1)
        # dH/dp to funkcja p
        # dH/dq to funkcja q

        terms = (ODETerm(lambda t, q, args: -jax.grad(H, argnums=1)(jnp.zeros_like(q), q)),
                 ODETerm(lambda t, p, args: jax.grad(H, argnums=0)(p, jnp.zeros_like(p)))
                 )
        #p,q
        y0 = (jnp.asarray([1.,0.]), jnp.asarray([0.,1.]) )
        t1 = 2*jnp.pi/4
        n=2**12
        dt = t1/n
        saveat = SaveAt(t1=True)
        solution = diffrax.diffeqsolve(terms,rat,0.0,t1,dt0=dt,y0=y0,saveat=saveat)
        p1,q1 = solution.ys
        self.assertTrue( np.allclose(p1,jnp.asarray([0.,-1.]), rtol=1e-4, atol=1e-4) )
        self.assertTrue( np.allclose(q1, jnp.asarray([1., 0.]), rtol=1e-4, atol=1e-4) )


    def test_circle_pot(self):
        def constrain(q):
            return jnp.sqrt(jnp.sum(q**2, keepdims=True))-1.

        rat = Rattle(nonlinear_solver=NewtonNonlinearSolver(rtol=1e-4, atol=1e-6),constrain=constrain)


        def H(p, q):

            return p @ p.T / 2. + q[1]



        terms = (ODETerm(lambda t, q, args: -jax.grad(H, argnums=1)(jnp.zeros_like(q), q)),
                 ODETerm(lambda t, p, args: jax.grad(H, argnums=0)(p, jnp.zeros_like(p)))
                 )
        #p,q
        y0 = (jnp.asarray([1.,0.]), jnp.asarray([0.,1.]) )
        t1 = 2*jnp.pi/4
        n=2**12
        dt = t1/n
        saveat = SaveAt(t1=True,dense=True)
        solution = diffrax.diffeqsolve(terms,rat,0.0,t1,dt0=dt,y0=y0,saveat=saveat)
        p1,q1 = solution.ys

    def test_circle_tree(self):

        class Q(NamedTuple):
            x:Array
            y:Array
        def constrain(q:Q):
            return jnp.sqrt(q.x**2+q.y**2)-1.

        rat = Rattle(nonlinear_solver=NewtonNonlinearSolver(rtol=1e-4, atol=1e-6),constrain=constrain)


        def H(p:Q, q:Q):
            del q
            return (p.x**2+p.y**2)/2.


        terms = (ODETerm(lambda t, q, args: (-jax.grad(H, argnums=1)(jtu.tree_map(jnp.zeros_like,q), q)**ω).ω),
                 ODETerm(lambda t, p, args: jax.grad(H, argnums=0)(p, jtu.tree_map(jnp.zeros_like,p)))
                 )


        #p,q
        y0 = (Q(x=1.,y=0.),Q(x=0.,y=1.))
        y0 = jtu.tree_map(jnp.asarray,y0)
        t1 = 2*jnp.pi/4
        n=2**12
        dt = t1/n
        saveat = SaveAt(t1=True, dense=True)


        #terms[0].vf(0.0, y0[0], None)

        solution = diffrax.diffeqsolve(terms,rat,0.0,t1,dt0=dt,y0=y0,saveat=saveat)
        p1,q1 = solution.ys
        self.assertAlmostEqual(p1.y, -1, places=4)

        t = jnp.linspace(0.0,t1,100)
        ps,qs = jax.vmap(solution.evaluate)(t)



if __name__ == '__main__':
    unittest.main()



import timeit

import jax
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
import seaborn as sns

from jax_chmc.kernels import fun_chmc


def make_chain(n=4):
    M = 3.0 * jnp.diag(jnp.ones(n))
    cm = fun_chmc(logdensity_fn=lambda q: -jnp.square(q).sum(),
                  sim_logdensity_fn=lambda q: -jnp.square(q).sum(),
                  con_fn=lambda q: q.sum(keepdims=True),
                  inverse_mass_matrix=M,
                  num_integration_steps=3,
                  step_size=0.3)
    return cm


def app():
    cm = make_chain()

    k = jax.random.PRNGKey(42)

    s0 = cm.init(jnp.zeros(4))
    s1, _ = cm.step(k, s0)

    ks = jax.random.split(k, 55048)
    _, qs = jax.lax.scan(lambda s, k: (cm.step(k, s)[0], s.position), s1, ks)

    sns.kdeplot(qs[5000:, :])
    plt.show()


def benchmark(few_large=True):
    """
    Apple M1 Pro  Sample mcmc: 0.14694534553905214 s ± 0.004017746250481866 s
        Lineax    Sample mcmc: 0.15854501789837627 s ± 0.0014483560794849987 s
        R**40     Sample mcmc: 0.20070723095312815 s ± 0.010753018901673742 s
        Li R**40  Sample mcmc: 0.2309575584294521 s ± 0.006085973709004624 s

    """
    n=40
    cm = make_chain(n)
    k = jax.random.PRNGKey(42)

    s0 = cm.init(jnp.zeros(n))
    s1, _ = cm.step(k, s0)

    if few_large:
        n_step = 256
        n = 4
    else:
        n_step = 2
        n = 4 * 128
    r = 16

    ks = jax.random.split(k, n_step)


    def sample():
        _, qs = jax.lax.scan(lambda s, k: (cm.step(k, s)[0], s.position), s1, ks)
        qs.block_until_ready()

    s = s0

    def _sample_py():
        s = s0
        for k in ks:
            s, q = cm.step(k, s)
            jax.block_until_ready(q)

    sample()
    t = timeit.repeat(sample, repeat=r, number=n, globals=dict(s=s))
    print(jax.devices())
    print(f'Sample mcmc: {np.mean(t) / r} s ± {2 * np.std(t) / np.sqrt(n * r)} s')


if __name__ == '__main__':
    benchmark()



import jax.numpy as jnp
import numpy as np


def conditional_mvn(loc, cov, a):
    """
    Conditional; distribution https://en.wikipedia.org/wiki/Multivariate_normal_distribution#Conditional_distributions
    :param loc: location
    :param cov: covaraince
    :param a: vector we condition on. every `np.nan` is considered unknown
    :return: loc and covariance
    """
    unknown_idx, = np.where(jnp.isnan(a))
    known_idx = np.setdiff1d(np.arange(len(a)), unknown_idx)
    mu1 = loc[unknown_idx]
    mu2 = loc[known_idx]
    sigma11 = cov[unknown_idx, :][:, unknown_idx]
    sigma22 = cov[known_idx, :][:, known_idx]
    sigma21 = cov[known_idx, :][:, unknown_idx]
    sigma12 = cov[unknown_idx, :][:, known_idx]

    mubar = mu1 - sigma12 @ np.linalg.solve(sigma22, a[known_idx] - mu2)
    sigmabar = sigma11 - sigma12 @ np.linalg.solve(sigma22, sigma21)
    return mubar, sigmabar


def constrained_mvn(loc, cov, A):
    """Compute loc and covariance os a mvn distribution constrained to  :math:`A@x=0`
    First we make a joint distribution od :math:`x` and the constraints.
    Next the constraints are marginalized by means of  ``conditional_mvn``.

    :param loc: lcoation
    :param cov: covariance matrix
    :param A:constrain vector
    :return:loc and covariance
    """
    A = np.atleast_2d(A)
    Aall = np.concatenate([np.eye(len(loc)), A], axis=0)
    loc_hat = Aall @ loc
    cov_hat = Aall @ cov @ Aall.T
    a = np.ones_like(loc_hat)[:, 0] + jnp.nan
    a[-1] = 0
    mubar, sigmabar = conditional_mvn(loc_hat, cov_hat, a)
    return mubar, sigmabar



from typing import Callable, NamedTuple, Any

import jax
import jax.numpy as jnp
from jax.scipy.linalg import solve_triangular, cholesky
from jaxtyping import PyTree, Array, Float, PRNGKeyArray

from jax_chmc.rattle import Rattle
from diffrax import NewtonNonlinearSolver
import jax.tree_util as jtu
from diffrax import ODETerm, SaveAt
from equinox.internal import ω
import diffrax


class CHMCState(NamedTuple):
    """State of the CHMC algorithm.

    The CHMC algorithm takes one position of the chain and returns another
    position. In order to make computations more efficient, we also store
    the current logdensity as well as the current gradient of the logdensity and the
    constraint jacobian.

    """
    position: PyTree
    # hamiltonian:float
    # logdensity: float
    # logdensity_grad: ArrayTree
    constrain_jac: Array


class CHMCInfo(NamedTuple):
    momentum: PyTree
    acceptance_rate: float
    is_accepted: bool
    is_divergent: bool
    energy: float
    proposal: Any
    num_integration_steps: int


class Mass:
    cholesky: Array
    inverse: Array

    def __init__(self, M: Array):
        self.cholesky = cholesky(M)
        self.inverse = solve_triangular(self.cholesky.T, solve_triangular(self.cholesky, jnp.eye(*M.shape), lower=True),
                                        lower=False)

    def compute_log_norm_const(self, dc: Array) -> Float:
        # https://math.stackexchange.com/questions/3155163/computing-the-pdf-of-a-low-rank-multivariate-normal

        D = dc @ self.inverse
        cholMhat = self.cholesky - D.T @ jnp.linalg.solve(D @ D.T, D @ self.cholesky)
        d = jnp.linalg.svd(cholMhat, compute_uv=False, hermitian=True)
        top_d, _ = jax.lax.top_k(d, dc.shape[1] - dc.shape[0])
        return jnp.sum(jnp.log(top_d))



class SamplingAlgorithm(NamedTuple):
    init: Callable
    step: Callable


def fun_chmc(
        logdensity_fn: Callable,  # H
        sim_logdensity_fn: Callable,  # hat H
        con_fn: Callable,
        step_size,  # h
        inverse_mass_matrix,  # M
        num_integration_steps,  # L
) -> SamplingAlgorithm:
    mass = Mass(inverse_mass_matrix)
    j_con_fun = jax.jacobian(con_fn)

    def generare_momentum(state: CHMCState, proposal_key):
        z = jax.random.normal(proposal_key, shape=state.position.shape)
        p0 = mass.cholesky @ z
        # projection
        dc = state.constrain_jac
        D = dc @ mass.inverse
        p0 = p0 - D.T @ jnp.linalg.solve(D @ D.T, D @ p0)

        return p0

    def init(position: PyTree) -> CHMCState:
        f, df = jax.value_and_grad(logdensity_fn)(position)
        jac = j_con_fun(position)

        return CHMCState(position=position,
                         # hamiltonian=0.,# hamiltonian be pendu
                         # logdensity_grad=df,
                         constrain_jac=jac
                         )

    def make_hamiltonian(logdensity_fn: Callable, sim=False):
        def hamiltonian(p: Array, q: Array):
            dc = j_con_fun(q)
            if sim:
                return 0.5 * p.T @ mass.inverse @ p - logdensity_fn(q)
            else:
                return 0.5 * p.T @ mass.inverse @ p + mass.compute_log_norm_const(dc) - logdensity_fn(q)

        return hamiltonian


    def kernel(
            rng_key: PRNGKeyArray,
            state: CHMCState,
    ) -> tuple[CHMCState, CHMCInfo]:
        proposal_key, accept_key = jax.random.split(rng_key)

        p0 = generare_momentum(state, proposal_key)

        target_H = make_hamiltonian(logdensity_fn, sim=False)

        pq0 = (p0, state.position)

        rat = Rattle(nonlinear_solver=NewtonNonlinearSolver(rtol=1e-4, atol=1e-6), constrain=con_fn)
        H = make_hamiltonian(sim_logdensity_fn, sim=True)
        terms = (ODETerm(lambda t, q, args: (-jax.grad(H, argnums=1)(jtu.tree_map(jnp.zeros_like, q), q) ** ω).ω),
                 ODETerm(lambda t, p, args: jax.grad(H, argnums=0)(p, jtu.tree_map(jnp.zeros_like, p))))
        saveat = SaveAt(t1=True)
        dt = 0.1
        t1 = num_integration_steps * dt
        solution = diffrax.diffeqsolve(terms, rat, 0.0, t1, dt0=dt, y0=pq0, saveat=saveat)
        pqL = (solution.ys[0][0], solution.ys[1][0])

        H0 = target_H(*pq0)
        H = target_H(*pqL)
        accept_p = jnp.minimum(1., jnp.exp(-(H - H0)))
        accept = jax.random.bernoulli(accept_key, accept_p)

        info = CHMCInfo(acceptance_rate=accept_p, is_accepted=accept,
                        proposal=pqL, momentum=pqL[0], is_divergent=None, energy=H,
                        num_integration_steps=num_integration_steps)

        # TODO improve performance
        new_state = jax.lax.cond(info.is_accepted, lambda: init(pqL[1]), lambda: state)

        return new_state, info

    return SamplingAlgorithm(init, kernel)



import unittest

import jax.numpy as jnp
import jax.random
import numpy as np

from jax_chmc import mvn
from jax_chmc import Mass
from jax_chmc import fun_chmc


class FunCHMCCase(unittest.TestCase):
    def test_something(self):
        logdens = lambda x: -(x ** 2).sum()
        condfun = lambda x: x.sum(keepdims=True)
        M = 3.0 * jnp.diag(jnp.ones(4))
        c = fun_chmc(logdensity_fn=logdens,
                     sim_logdensity_fn=logdens,
                     con_fn=condfun,
                     step_size=0.1,
                     inverse_mass_matrix=M,
                     num_integration_steps=2
                     )

        s0 = c.init(jnp.ones(4))
        k = jax.random.PRNGKey(52)
        s1 = c.step(k, s0)
        ...

    def test_mvn(self):
        M = 3.0 * jnp.diag(jnp.ones(4))
        k = jax.random.PRNGKey(42)

        init_fn, step_fn = fun_chmc(logdensity_fn=lambda q: -jnp.square(q).sum(),
                                    sim_logdensity_fn=lambda q: -jnp.square(q).sum(),
                                    con_fn=lambda q: q[:2].sum(keepdims=True),
                                    step_size=0.1,
                                    inverse_mass_matrix=M,
                                    num_integration_steps=2
                                    )

        s0 = init_fn(jnp.zeros(4))

        s1, _ = step_fn(k, s0)

        ks = jax.random.split(k, 55048)

        def step(state, k):
            new_state, info = step_fn(k, state)
            return new_state, (new_state, info)

        state, (states, infos) = jax.lax.scan(step, s1, ks)

        cov = 0.5 * np.eye(s0.position.shape[0])
        loc = np.zeros((4, 1))
        A = jnp.asarray([1, 1, 0, 0])
        mubar, sigmabar = mvn.constrained_mvn(loc, cov, A)
        qs = states.position
        sampledsigma = np.cov(qs[2000:, :].T)
        self.assertTrue(np.allclose(sampledsigma, sigmabar, atol=0.05))


class MassCase(unittest.TestCase):

    def test_mass(self):
        M = 3.1 * jnp.diag(jnp.ones(4))
        m = Mass(M)


if __name__ == '__main__':
    unittest.main()



from typing import Tuple, Callable, NamedTuple

import diffrax
import jax
import jax.numpy as jnp
import jax.tree_util as jtu
from diffrax import AbstractImplicitSolver
from diffrax import AbstractTerm
from diffrax.custom_types import Bool, DenseInfo, PyTree, Scalar, Array
from diffrax.local_interpolation import LocalLinearInterpolation
from diffrax.solution import RESULTS
from equinox.internal import ω

_ErrorEstimate = None
_SolverState = None


class RattleVars(NamedTuple):
    p_1_2: PyTree  # Midpoint momentum
    q_1: PyTree  # Midpoint position
    p_1: PyTree  # final momentum
    lam: PyTree  # Midpoint Lagrange multiplier (state)
    mu: PyTree  # final Lagrange multiplier (momentum)


class Rattle(AbstractImplicitSolver):
    """ Rattle method.

    Symplectic method. Does not support adaptive step sizing. Uses 1st order local
    linear interpolation for dense/ts output.
     dv = f(t,w(t)) dt
     dw = g(t,v(t))dt
     Hamiltonian
     H = V(p) + U(q)
     dp = -∂H/∂q dt
     dq = ∂H/∂p dt

    """

    term_structure = (AbstractTerm, AbstractTerm)
    interpolation_cls = LocalLinearInterpolation
    constrain: Callable = None  # Fix TypeError: non-default argument 'constrain' follows default argument

    def order(self, terms):
        return 2

    def init(self, terms: Tuple[AbstractTerm, AbstractTerm], t0: Scalar, t1: Scalar, y0: PyTree,
             args: PyTree, ) -> _SolverState:
        return None

    def step(self, terms: Tuple[AbstractTerm, AbstractTerm], t0: Scalar, t1: Scalar, y0: Tuple[PyTree, PyTree],
             args: PyTree, solver_state: _SolverState, made_jump: Bool, ) -> Tuple[
        Tuple[PyTree, PyTree], _ErrorEstimate, DenseInfo, _SolverState, RESULTS]:
        del solver_state, made_jump

        term_1, term_2 = terms
        y0_1, y0_2 = y0  # p, q

        midpoint = (t1 + t0) / 2

        # p
        control1_half_1 = term_1.contr(t0, midpoint)
        control1_half_2 = term_1.contr(midpoint, t1)

        # q
        control2_half_1 = term_2.contr(t0, midpoint)
        control2_half_2 = term_2.contr(midpoint, t1)

        p0 = y0_1
        q0 = y0_2

        p0, q0 = y0

        def eq(x: RattleVars, args=None):
            _, vjp_fun = jax.vjp(self.constrain, q0)
            _, vjp_fun_mu = jax.vjp(self.constrain, x.q_1)

            zero = ((p0 ** ω - control1_half_1 * (vjp_fun(x.lam)[0]) ** ω + term_1.vf_prod(t0, q0, args,
                                                                                           control1_half_1) ** ω - x.p_1_2 ** ω).ω,
                    (q0 ** ω + term_2.vf_prod(t0, x.p_1_2, args, control2_half_1) ** ω + term_2.vf_prod(midpoint,
                                                                                                        x.p_1_2, args,
                                                                                                        control2_half_2) ** ω - x.q_1 ** ω).ω,
                    self.constrain(x.q_1), (
                            x.p_1_2 ** ω + term_1.vf_prod(midpoint, x.q_1, args, control1_half_2) ** ω - (
                            control1_half_2 * vjp_fun_mu(x.mu)[0] ** ω) - x.p_1 ** ω).ω,
                    jax.jvp(self.constrain, (x.q_1,), (term_2.vf(t1, x.p_1, args),))[1])
            return zero

        cs = jax.eval_shape(self.constrain, q0)

        init_vars = RattleVars(p_1_2=p0, q_1=(q0 ** ω * 2).ω, p_1=p0,
                               lam=jtu.tree_map(lambda cs: jnp.zeros(cs.shape, cs.dtype), cs),
                               mu=jtu.tree_map(lambda cs: jnp.zeros(cs.shape, cs.dtype), cs))

        sol = self.nonlinear_solver(eq, init_vars, None)

        y1 = (sol.root.p_1, sol.root.q_1)
        dense_info = dict(y0=y0, y1=y1)
        return y1, None, dense_info, None, RESULTS.successful

    def func(self, terms: Tuple[AbstractTerm, AbstractTerm], t0: Scalar, y0: Tuple[PyTree, PyTree], args: PyTree) -> \
            Tuple[PyTree, PyTree]:
        term_1, term_2 = terms
        y0_1, y0_2 = y0
        f1 = term_1.func(t0, y0_2, args)
        f2 = term_2.func(t0, y0_1, args)
        return (f1, f2)


if __name__ == '__main__':
    from diffrax import AbstractImplicitSolver, SaveAt, ODETerm, NewtonNonlinearSolver


    class Q(NamedTuple):
        x: Array
        y: Array


    def constrain(q: Q):
        return dict(c=jnp.sqrt(q.x ** 2 + q.y ** 2) - 1.)  # check pytree constrain


    rat = Rattle(nonlinear_solver=NewtonNonlinearSolver(rtol=1e-4, atol=1e-6), constrain=constrain)


    def H(p: Q, q: Q):
        del q
        return (p.x ** 2 + p.y ** 2) / 2.


    def H(p: Q, q: Q):
        return (p.x ** 2 + p.y ** 2) / 2. + q.y


    terms = (ODETerm(lambda t, q, args: (-jax.grad(H, argnums=1)(jtu.tree_map(jnp.zeros_like, q), q) ** ω).ω),
             ODETerm(lambda t, p, args: jax.grad(H, argnums=0)(p, jtu.tree_map(jnp.zeros_like, p))))

    # p,q
    y0 = (Q(x=1., y=0.), Q(x=0., y=1.))
    y0 = jtu.tree_map(jnp.asarray, y0)
    t1 = 2 * jnp.pi / 4
    n = 2 ** 12
    dt = t1 / n
    saveat = SaveAt(t1=True, dense=True)

    solution = diffrax.diffeqsolve(terms, rat, 0.0, t1, dt0=dt, y0=y0, saveat=saveat)
    p1, q1 = solution.ys

    t = jnp.linspace(0.0, t1, 100)
    ps, qs = jax.vmap(solution.evaluate)(t)
    import matplotlib.pyplot as plt

    plt.plot(qs.x, qs.y)

    plt.gca().set_aspect('equal')
    plt.figure()
    plt.plot(t, jax.vmap(lambda q: jnp.sqrt(q.x ** 2 + q.y ** 2))(ps))
    plt.show()
